/** @file oathtool_cmd.h
 *  @brief The header file for the command line option parser
 *  generated by GNU Gengetopt version 2.23
 *  http://www.gnu.org/software/gengetopt.
 *  DO NOT modify this file, since it can be overwritten
 *  @author GNU Gengetopt */

#ifndef OATHTOOL_CMD_H
#define OATHTOOL_CMD_H

/* If we use autoconf.  */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h> /* for FILE */

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#ifndef CMDLINE_PARSER_PACKAGE
/** @brief the program name (used for printing errors) */
#define CMDLINE_PARSER_PACKAGE "oathtool"
#endif

#ifndef CMDLINE_PARSER_PACKAGE_NAME
/** @brief the complete program name (used for help and version) */
#define CMDLINE_PARSER_PACKAGE_NAME "oathtool"
#endif

#ifndef CMDLINE_PARSER_VERSION
/** @brief the program version */
#define CMDLINE_PARSER_VERSION VERSION
#endif

/** @brief Where the command line options are stored */
struct gengetopt_args_info
{
  const char *help_help; /**< @brief Print help and exit help description.  */
  const char *version_help; /**< @brief Print version and exit help description.  */
  int hotp_flag;	/**< @brief use event-based HOTP mode (default=on).  */
  const char *hotp_help; /**< @brief use event-based HOTP mode help description.  */
  char * totp_arg;	/**< @brief use time-variant TOTP mode (values \"SHA1\", \"SHA256\", or \"SHA512\") (default='SHA1').  */
  char * totp_orig;	/**< @brief use time-variant TOTP mode (values \"SHA1\", \"SHA256\", or \"SHA512\") original value given at command line.  */
  const char *totp_help; /**< @brief use time-variant TOTP mode (values \"SHA1\", \"SHA256\", or \"SHA512\") help description.  */
  int base32_flag;	/**< @brief use base32 encoding of KEY instead of hex (default=off).  */
  const char *base32_help; /**< @brief use base32 encoding of KEY instead of hex help description.  */
  #if defined(HAVE_UNSIGNED_LONG_LONG) || defined(HAVE_UNSIGNED_LONG_LONG_INT)
  unsigned long long int counter_arg;	/**< @brief HOTP counter value.  */
  #else
  long counter_arg;	/**< @brief HOTP counter value.  */
  #endif
  char * counter_orig;	/**< @brief HOTP counter value original value given at command line.  */
  const char *counter_help; /**< @brief HOTP counter value help description.  */
  char * time_step_size_arg;	/**< @brief TOTP time-step duration (default='30s').  */
  char * time_step_size_orig;	/**< @brief TOTP time-step duration original value given at command line.  */
  const char *time_step_size_help; /**< @brief TOTP time-step duration help description.  */
  char * start_time_arg;	/**< @brief when to start counting time steps for TOTP (default='1970-01-01 00:00:00 UTC').  */
  char * start_time_orig;	/**< @brief when to start counting time steps for TOTP original value given at command line.  */
  const char *start_time_help; /**< @brief when to start counting time steps for TOTP help description.  */
  char * now_arg;	/**< @brief use this time as current time for TOTP (default='now').  */
  char * now_orig;	/**< @brief use this time as current time for TOTP original value given at command line.  */
  const char *now_help; /**< @brief use this time as current time for TOTP help description.  */
  int digits_arg;	/**< @brief number of digits in one-time password.  */
  char * digits_orig;	/**< @brief number of digits in one-time password original value given at command line.  */
  const char *digits_help; /**< @brief number of digits in one-time password help description.  */
  int window_arg;	/**< @brief number of additional OTPs to generate or validate against.  */
  char * window_orig;	/**< @brief number of additional OTPs to generate or validate against original value given at command line.  */
  const char *window_help; /**< @brief number of additional OTPs to generate or validate against help description.  */
  int verbose_flag;	/**< @brief explain what is being done (default=off).  */
  const char *verbose_help; /**< @brief explain what is being done help description.  */
  
  unsigned int help_given ;	/**< @brief Whether help was given.  */
  unsigned int version_given ;	/**< @brief Whether version was given.  */
  unsigned int hotp_given ;	/**< @brief Whether hotp was given.  */
  unsigned int totp_given ;	/**< @brief Whether totp was given.  */
  unsigned int base32_given ;	/**< @brief Whether base32 was given.  */
  unsigned int counter_given ;	/**< @brief Whether counter was given.  */
  unsigned int time_step_size_given ;	/**< @brief Whether time-step-size was given.  */
  unsigned int start_time_given ;	/**< @brief Whether start-time was given.  */
  unsigned int now_given ;	/**< @brief Whether now was given.  */
  unsigned int digits_given ;	/**< @brief Whether digits was given.  */
  unsigned int window_given ;	/**< @brief Whether window was given.  */
  unsigned int verbose_given ;	/**< @brief Whether verbose was given.  */

  char **inputs ; /**< @brief unnamed options (options without names) */
  unsigned inputs_num ; /**< @brief unnamed options number */
} ;

/** @brief The additional parameters to pass to parser functions */
struct cmdline_parser_params
{
  int override; /**< @brief whether to override possibly already present options (default 0) */
  int initialize; /**< @brief whether to initialize the option structure gengetopt_args_info (default 1) */
  int check_required; /**< @brief whether to check that all required options were provided (default 1) */
  int check_ambiguity; /**< @brief whether to check for options already specified in the option structure gengetopt_args_info (default 0) */
  int print_errors; /**< @brief whether getopt_long should print an error message for a bad option (default 1) */
} ;

/** @brief the purpose string of the program */
extern const char *gengetopt_args_info_purpose;
/** @brief the usage string of the program */
extern const char *gengetopt_args_info_usage;
/** @brief the description string of the program */
extern const char *gengetopt_args_info_description;
/** @brief all the lines making the help output */
extern const char *gengetopt_args_info_help[];

/**
 * The command line parser
 * @param argc the number of command line options
 * @param argv the command line options
 * @param args_info the structure where option information will be stored
 * @return 0 if everything went fine, NON 0 if an error took place
 */
int cmdline_parser (int argc, char **argv,
  struct gengetopt_args_info *args_info);

/**
 * The command line parser (version with additional parameters - deprecated)
 * @param argc the number of command line options
 * @param argv the command line options
 * @param args_info the structure where option information will be stored
 * @param override whether to override possibly already present options
 * @param initialize whether to initialize the option structure my_args_info
 * @param check_required whether to check that all required options were provided
 * @return 0 if everything went fine, NON 0 if an error took place
 * @deprecated use cmdline_parser_ext() instead
 */
int cmdline_parser2 (int argc, char **argv,
  struct gengetopt_args_info *args_info,
  int override, int initialize, int check_required);

/**
 * The command line parser (version with additional parameters)
 * @param argc the number of command line options
 * @param argv the command line options
 * @param args_info the structure where option information will be stored
 * @param params additional parameters for the parser
 * @return 0 if everything went fine, NON 0 if an error took place
 */
int cmdline_parser_ext (int argc, char **argv,
  struct gengetopt_args_info *args_info,
  struct cmdline_parser_params *params);

/**
 * Save the contents of the option struct into an already open FILE stream.
 * @param outfile the stream where to dump options
 * @param args_info the option struct to dump
 * @return 0 if everything went fine, NON 0 if an error took place
 */
int cmdline_parser_dump(FILE *outfile,
  struct gengetopt_args_info *args_info);

/**
 * Save the contents of the option struct into a (text) file.
 * This file can be read by the config file parser (if generated by gengetopt)
 * @param filename the file where to save
 * @param args_info the option struct to save
 * @return 0 if everything went fine, NON 0 if an error took place
 */
int cmdline_parser_file_save(const char *filename,
  struct gengetopt_args_info *args_info);

/**
 * Print the help
 */
void cmdline_parser_print_help(void);
/**
 * Print the version
 */
void cmdline_parser_print_version(void);

/**
 * Initializes all the fields a cmdline_parser_params structure 
 * to their default values
 * @param params the structure to initialize
 */
void cmdline_parser_params_init(struct cmdline_parser_params *params);

/**
 * Allocates dynamically a cmdline_parser_params structure and initializes
 * all its fields to their default values
 * @return the created and initialized cmdline_parser_params structure
 */
struct cmdline_parser_params *cmdline_parser_params_create(void);

/**
 * Initializes the passed gengetopt_args_info structure's fields
 * (also set default values for options that have a default)
 * @param args_info the structure to initialize
 */
void cmdline_parser_init (struct gengetopt_args_info *args_info);
/**
 * Deallocates the string fields of the gengetopt_args_info structure
 * (but does not deallocate the structure itself)
 * @param args_info the structure to deallocate
 */
void cmdline_parser_free (struct gengetopt_args_info *args_info);

/**
 * Checks that all the required options were specified
 * @param args_info the structure to check
 * @param prog_name the name of the program that will be used to print
 *   possible errors
 * @return
 */
int cmdline_parser_required (struct gengetopt_args_info *args_info,
  const char *prog_name);


#ifdef __cplusplus
}
#endif /* __cplusplus */
#endif /* OATHTOOL_CMD_H */
